{% extends "sphinxdoc/layout.html" %}

{#
{% block header %}

<div style="background-color: white; text-align: left; padding: 10px 10px 15px 15px">
<img class="logo" src="{{ pathto('_static/pyramid.png', 1)}}" alt="Logo"/>
</div>
{% endblock %}
#}
{%- block sidebarsourcelink %}{%- endblock %}
